import geojson
import json
import pendulum
import requests
import shapely
import subprocess
import time
import typer
from geoalchemy2.shape import to_shape
from logger import Logger
from settings import Settings
from shapely.geometry import shape
from sql import crud, models, database, schemas
import urllib.request
from typing_extensions import Annotated


settings = Settings()
logging = Logger(__name__)
log = logging.log

app = typer.Typer()


@app.command()
def create_db():
    db = database.Database()
    # schemas.Categories.__table__.create(db.engine)
    # schemas.Points.__table__.create(db.engine)
    # schemas.VigiloCategories.__table__.create(db.engine)
    # schemas.VigiloObservations.__table__.create(db.engine)
    # schemas.UnMetre.__table__.create(db.engine)
    # schemas.Onisr_2021_Processed.__table__.create(db.engine)
    # schemas.Onisr_2022_Processed.__table__.create(db.engine)


@app.command()
def populate_fub_data_by_file(
    file: str, categorie: str, force: bool = typer.Option(False, help="Force update")
):
    if not crud.is_categorie_exist(categorie):
        log.warning(f"Categorie {categorie} not exist. Creating")
        crud.create_categorie(categorie_name=categorie)
    categories = crud.get_categorie(categorie)
    with open(f"{file}", "r") as json_file:
        data = json.load(json_file)
        for record in data["features"]:
            point = models.PointsCreate(
                id_categories=categories.id,
                geolocalisation=f"POINT({record['geometry']['coordinates'][0]} {record['geometry']['coordinates'][1]})",
            )
            crud.create_point(point)
            log.debug(f"Adding {record['geometry']['coordinates']}")


@app.command()
def populate_vigilo_ctg():
    url = "https://raw.githubusercontent.com/jesuisundesdeux/vigilo-conf/main/main/categorielist.json"
    req = requests.get(url=url)
    data_req = req.json()
    for data in data_req:
        categorie = models.VigiloCategories(id=data["catid"], name=data["catname"])
        if not crud.is_vigilo_categories_exists(categorie):
            log.warning(f"Vigilo categorie \"{data['catname']}\" not exist. Creating")
            crud.create_vigilo_categories(categorie)


@app.command()
def populate_full_vigilo_observation():
    url = "https://vigilo.placeauvelo-nantes.fr//get_issues.php?scope=44_nantesmetropole&format=json&count=40000&approved=1"
    req = requests.get(url=url)
    data_req = req.json()
    for data in data_req:
        observation = models.VigiloObservations(
            id=data["token"],
            vigilo_categories_id=data["categorie"],
            datetime=data["time"],
            geolocalisation=f"POINT({data['coordinates_lat']} {data['coordinates_lon']})",
            comment=data["comment"],
        )
        if not crud.is_vigilo_observation_exists(observation):
            log.warning(f"Observation {data['token']} don't exist. Creating")
            crud.create_vigilo_observation(observation)


@app.command()
def populate_vigilo_observation(
    days: Annotated[int, typer.Option(help="Number of days to get occurences")] = 14
):
    url = f"https://vigilo.placeauvelo-nantes.fr//get_issues.php?scope=44_nantesmetropole&format=json&count=40000&approved=1&since={days}&since_unit=day"
    req = requests.get(url=url)
    data_req = req.json()
    for data in data_req:
        observation = models.VigiloObservations(
            id=data["token"],
            vigilo_categories_id=data["categorie"],
            datetime=data["time"],
            geolocalisation=f"POINT({data['coordinates_lat']} {data['coordinates_lon']})",
            comment=data["comment"],
        )
        if not crud.is_vigilo_observation_exists(observation):
            log.warning(f"Observation {data['token']} don't exist. Creating")
            crud.create_vigilo_observation(observation)
        else:
            log.debug(f"Observation {data['token']} exist. Pass")


@app.command()
def populate_geojson():
    geojson = [
        {
            "name": "1metre",
            "url": "https://nantes.1metre.fr/api/export/events?bbox=-1.3020903353663267,47.13226606037135,-1.763268432053864,47.32918298624632&fmt=geojson",
            "path": "/var/www/opendata_nm/www/nantes_unmetre.geojson",
        },
        {
            "name": "Nouveaux axes Nantes",
            "url": "https://data.nantesmetropole.fr/api/explore/v2.1/catalog/datasets/244400404_schema-directeur-infrastructures-cyclables-nantes-metropole/exports/geojson?lang=fr&timezone=Europe%2FBerlin",
            "path": "/var/www/opendata_nm/www/schema-directeur-infrastructures-cyclables-nantes-metropole.geojson",
        },
        {
            "name": "Pistes cyclables actuelles Nantes",
            "url": "https://data.nantesmetropole.fr/api/explore/v2.1/catalog/datasets/244400404_amenagements-cyclables-nantes-metropole-enrichis/exports/geojson?lang=fr&timezone=Europe%2FBerlin",
            "path": "/var/www/opendata_nm/www/pistes_cyclables_nantes.geojson",
        },
    ]
    for data in geojson:
        log.info(f"Downloading new version of {data['name']} file")
        headers = {"Accept": "application/json"}
        urllib.request.urlretrieve(data["url"], data["path"])
        log.info(f"New file for {data['name']} correctly downloaded")


@app.command()
def import_unmetre():
    import gdaltools

    geojson = [
        {
            "name": "1metre",
            "url": "https://nantes.1metre.fr/api/export/events?bbox=-1.3020903353663267,47.13226606037135,-1.763268432053864,47.32918298624632&fmt=geojson",
            "path": "/var/www/opendata_nm/www/nantes_unmetre.geojson",
        }
    ]
    for data in geojson:
        log.info(f"Downloading new version of {data['name']} file")
        headers = {"Accept": "application/json"}
        urllib.request.urlretrieve(data["url"], data["path"])
        log.info(f"New file for {data['name']} correctly downloaded")

    log.info("Drop un_metre database")
    crud.drop_unmetre()
    db = settings.db
    ogr = gdaltools.ogr2ogr()
    ogr.set_encoding("UTF-8")
    log.info("Setting")
    ogr.set_input("/var/www/opendata_nm/www/nantes_unmetre.geojson")
    conn = gdaltools.PgConnectionString(
        host=db["host"],
        port=db["port"],
        dbname=db["dbname"],
        user=db["user"],
        password=db["password"],
    )
    ogr.set_output(conn, table_name="un_metre")
    log.info("OGR Execute")
    ogr.execute()


@app.command()
def process_raw_onisr_value():
    for data in crud.get_raw_onisr_value():
        pendulum.set_locale("fr")
        dt = pendulum.parse(
            f"{int(data[8])}-{int(data[7])}-{int(data[6])} {data[9]}",
            strict=False,
            tz="Europe/Paris",
        )
        onisr_processed = models.OnisrProcessedCreate(
            accident_id=data[0],
            gravite=data[1],
            obstacle_mobile=data[2],
            categorie_vehicule=data[3],
            numero_vehicule=data[4],
            intersection=data[5],
            datetime=dt.to_iso8601_string(),
            geolocalisation=f"POINT({data[11]} {data[10]})",
        )
        crud.add_onisr_processed(onisr_processed)


@app.command()
def get_all_streets():
    import overpass

    communes = crud.get_all_communes()
    log.info("Get all cities")
    insee_code = []
    for commune in communes:
        insee = int(commune.identifiant_insee)
        insee_code.append(insee)
        name = commune.toponyme
        log.info(f"Get all road in {name} ({insee})")
        api = overpass.API()
        query = f"""
            area["ref:INSEE"="{insee}"];
            (
            way["highway"~"primary|secondary|tertiary|unclassified|residential|primary_link|secondary_link|tertiary_link|living_street|track|path|cycleway"]["bicycle"!~"no"]["access"!~"no"](area);
            way["biGycle"="yes"]["access"!~"no"](area);
            );
        """
        log.info("Run query")
        r = api.get(query, verbosity="geom", responseformat="geojson")
        log.info(f"Write geojson /var/www/opendata_nm/www/streets/{insee}.geojson")
        with open(f"/var/www/opendata_nm/www/streets/{insee}.geojson", mode="w") as f:
            geojson.dump(r, f)
        log.info(f"Geojson writen. Wait 10 seconds")
        time.sleep(10)
        log.info("Sleep well. Next city !")
    log.info("And of getting all cities")


@app.command()
def analyze_unmetre_streets(
    onlyscore: Annotated[bool, typer.Option(help="Run only score calculation")] = False,
    reset: Annotated[bool, typer.Option(help="Reset count")] = False,
):
    category = "unmetre"
    log.info("Get all cities")
    communes = crud.get_all_communes()
    log.info(f"Analyze streets with {category} data")
    data_categories = [
        {
            "type": f"{category}_legal",
            "where": "WHERE distance_overtaker >= 1",
            "score": 0,
        },
        {
            "type": f"{category}_illegal",
            "where": "WHERE distance_overtaker > 0.5 AND distance_overtaker < 1",
            "score": 1,
        },
        {
            "type": f"{category}_dangerous",
            "where": "WHERE distance_overtaker <= 0.5",
            "score": 2,
        },
    ]
    if not onlyscore:
        for data_category in data_categories:
            log.info(f"Get all points for {data_category['type']}")
            un_metre_datas = crud.get_unmetre_points(data_category["where"])
            points_list = []
            for un_metre_data in un_metre_datas:
                multipoint = shapely.from_wkb(un_metre_data.wkb_geometry)
                point = shapely.get_geometry(multipoint, 0)
                points_list.append(point)
            if reset:
                reset_streets(
                    category=category,
                    data_type=data_category["type"],
                    communes=communes,
                )
            analyze_streets(
                category=category,
                data_type=data_category["type"],
                points=points_list,
                communes=communes,
            )

    for data_category in data_categories:
        score_streets(
            category=category,
            data_type=data_category["type"],
            category_score=data_category["score"],
            communes=communes,
            reset=reset,
        )
    log.info(f"End of analyze streets with {category} data")


@app.command()
def analyze_onisr_streets(
    onlyscore: Annotated[bool, typer.Option(help="Run only score calculation")] = False,
    reset: Annotated[bool, typer.Option(help="Reset count")] = False,
):
    log.info("Get all cities")
    communes = crud.get_all_communes()
    category = "onisr"
    data_categories = [
        {"type": f"{category}_2021", "score": 0.5},
        {"type": f"{category}_2022", "score": 1},
    ]
    if not onlyscore:
        points_list = []
        for data_category in data_categories:
            log.info(f"Get all points for {data_category['type']}")
            if data_category["type"] == f"{category}_2021":
                log.info(f"Use {category}_2021")
                onisr_datas = crud.get_all_onisr_2021()
            elif data_category["type"] == f"{category}_2022":
                log.info(f"Use {category}_2022")
                onisr_datas = crud.get_all_onisr_2022()
            for onisr_data in onisr_datas:
                point = shapely.get_geometry(to_shape(onisr_data.geolocalisation), 0)
                points_list.append(point)
            log.debug(f"{data_category['type']} have {len(points_list)} points")
            if reset:
                reset_streets(
                    category=category,
                    data_type=data_category["type"],
                    communes=communes,
                )
            analyze_streets(
                category=category,
                data_type=data_category["type"],
                points=points_list,
                communes=communes,
                precision=1e-4,
            )

    for data_category in data_categories:
        score_streets(
            category=category,
            data_type=data_category["type"],
            category_score=data_category["score"],
            communes=communes,
            reset=reset,
        )


@app.command()
def analyze_vigilo_streets(
    onlyscore: Annotated[bool, typer.Option(help="Run only score calculation")] = False,
    reset: Annotated[bool, typer.Option(help="Reset count")] = False,
):
    log.info("Get all cities")
    communes = crud.get_all_communes()
    category = "vigilo"
    data_categories = [
        {"type": f"{category}_genant", "score": 2, "category_id": 2},
        {"type": f"{category}_incident", "score": 2, "category_id": 9},
        {"type": f"{category}_incivilite", "score": 1, "category_id": 7},
    ]
    if not onlyscore:
        points_list = []
        for data_category in data_categories:
            log.info(f"Get all points for {data_category['type']}")
            vigilo_datas = crud.get_vigilo(data_category["category_id"])
            for vigilo in vigilo_datas:
                point = shapely.get_geometry(to_shape(vigilo.geolocalisation), 0)
                points_list.append(point)
            log.debug(f"{data_category['type']} have {len(points_list)} points")
            if reset:
                reset_streets(
                    category=category,
                    data_type=data_category["type"],
                    communes=communes,
                )
            analyze_streets(
                category=category,
                data_type=data_category["type"],
                points=points_list,
                communes=communes,
                precision=1e-4,
            )
    for data_category in data_categories:
        score_streets(
            category=category,
            data_type=data_category["type"],
            category_score=data_category["score"],
            communes=communes,
            reset=reset,
        )


def analyze_streets(
    category: str, data_type: str, points: list, communes, precision: float = 1e-7
):
    log.info(f"Process category {data_type}")
    for commune in communes:
        insee = int(commune.identifiant_insee)
        name = commune.toponyme
        log.info(f"Load geojson for {name} ({insee})")
        with open(f"/var/www/opendata_nm/www/streets/{insee}.geojson", mode="r") as f:
            data = geojson.load(f)
        for feature in data["features"]:
            if feature["properties"].get(f"{category}_all") is None:
                feature["properties"][f"{category}_all"] = 0
            linestring = shape(feature["geometry"])
            for point in points:
                if linestring.distance(point) < precision:
                    if feature["properties"].get("name") is not None:
                        street_name = feature["properties"]["name"]
                    else:
                        street_name = "unknown"

                    if feature["properties"].get(f"{data_type}") is None:
                        log.info(
                            f"[{data_type}] Point {point} found in a new street ({street_name})"
                        )
                        feature["properties"][f"{data_type}"] = 1
                    else:
                        feature["properties"][f"{data_type}"] += 1
                        log.info(
                            f"[{data_type}]Point {point} found in an already known street ({street_name})."
                        )

                    log.debug(
                        "Adding one to {}".format(
                            feature["properties"][f"{category}_all"]
                        )
                    )
                    feature["properties"][f"{category}_all"] += 1

        with open(f"/var/www/opendata_nm/www/streets/{insee}.geojson", mode="w") as f:
            geojson.dump(data, f)
        log.info(f"Check streets complete for {data_type}")


def reset_streets(category: str, data_type: str, communes):
    log.info(f"Process category {data_type}")
    for commune in communes:
        insee = int(commune.identifiant_insee)
        name = commune.toponyme
        log.info(f"Load geojson for {name} ({insee})")
        with open(f"/var/www/opendata_nm/www/streets/{insee}.geojson", mode="r") as f:
            data = geojson.load(f)
        for feature in data["features"]:
            feature["properties"][f"{category}_all"] = 0
            feature["properties"][f"{data_type}"] = 0
        with open(f"/var/www/opendata_nm/www/streets/{insee}.geojson", mode="w") as f:
            geojson.dump(data, f)
        log.info(f"Check streets complete for {data_type}")


def score_streets(
    category: str, data_type: str, category_score: int, communes, reset: bool = False
):
    if category_score == 0:
        log.debug("category_score is 0")
        return
    for commune in communes:
        insee = int(commune.identifiant_insee)
        name = commune.toponyme
        log.info(f"[{data_type}] Load geojson for {name} ({insee})")
        with open(f"/var/www/opendata_nm/www/streets/{insee}.geojson", mode="r") as f:
            data = geojson.load(f)
        for feature in data["features"]:
            score = 0
            if reset:
                feature["properties"][f"{category}_score"] = 0
            if feature["properties"].get(f"{category}_score") is None:
                feature["properties"][f"{category}_score"] = 0
            if (
                feature["properties"].get(f"{data_type}") is not None
                and feature["properties"][f"{category}_all"] != 0
            ):
                score = (
                    feature["properties"][f"{data_type}"]
                    / feature["properties"][f"{category}_all"]
                    * 100
                    * category_score
                )
                # if feature["properties"].get("name") is not None and feature["properties"]["name"] == "Rue de Strasbourg":
                #    print(score)
                #    print(category_score)
                #    print(feature["properties"][f"{data_type}"])
                #    print(feature["properties"][f"{category}_all"])
                # log.debug("In {} ctg there are {} with total of {}. Score of {}".format(data_type, feature["properties"][f"{data_type}"], feature["properties"][f"{category}_all"], category_score))
            if score > 0:
                feature["properties"][f"{category}_score"] += score
        with open(f"/var/www/opendata_nm/www/streets/{insee}.geojson", mode="w") as f:
            geojson.dump(data, f)
    log.info("Process complete")


@app.command()
def calculate_score_streets(
    reset: Annotated[bool, typer.Option(help="Reset result")] = False,
):
    data_categories = [
        {"score_name": "onisr_score", "ponderate": 0.30},
        {"score_name": "unmetre_score", "ponderate": 0.60},
        {"score_name": "vigilo_score", "ponderate": 0.10},
    ]
    log.info("Get all cities")
    communes = crud.get_all_communes()
    for commune in communes:
        insee = int(commune.identifiant_insee)
        name = commune.toponyme
        log.info(f"Load geojson for {name} ({insee})")
        with open(f"/var/www/opendata_nm/www/streets/{insee}.geojson", mode="r") as f:
            data = geojson.load(f)
        for feature in data["features"]:
            score = 0
            if feature["properties"].get("score") is None or reset:
                feature["properties"]["score"] = 0
            for data_category in data_categories:
                if feature["properties"].get(data_category["score_name"]) is None:
                    feature["properties"][data_category["score_name"]] = 0
                score += (
                    feature["properties"][data_category["score_name"]]
                    * data_category["ponderate"]
                )
            if score < 0:
                feature["properties"]["score"] = -1
            else:
                feature["properties"]["score"] = score
            if feature["properties"]["score"] == 0:
                feature["properties"]["score"] = -1
        with open(f"/var/www/opendata_nm/www/streets/{insee}.geojson", mode="w") as f:
            geojson.dump(data, f)
    log.info("Process complete")


if __name__ == "__main__":
    app()
